--/run print(string.format("Unhittable at 102.4%% - you have %.2f%%", GetDodgeChance() + GetBlockChance() + GetParryChance() + 5 + 1/(0.0625 + 0.956/(GetCombatRating(CR_DEFENSE_SKILL)/4.91850*0.04))))

local Unhittable = LibStub("AceAddon-3.0"):NewAddon("Broker_Unhittable", "AceEvent-3.0", "AceBucket-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Unhittable")

local PlayerLevel = UnitLevel("player")
local PlayerClass = select(2, UnitClass("player"))

local db
local ldb

local defaults = {
	profile = {
		avoidColor = {0, 1, 0, 1},
		drColor = {0, .5, 1, 1},
		armorColor = {1, 1, 1, 1},
		headerColor = {0, .5, 1, 1},
		labelColor = {0, 1, 0, 1},
		valueColor = {1, 1, 1, 1},
		brokerLabel = 0,
		showFuBarAV = true,
		showFuBarAVBlock = true,
		showFuBarDR = true,
		showFuBarArmor = false,
		displayPrecisionFuBar = 2,
		showTooltipCombat = true,
		showTooltipArmor = true,
		displayPrecisionTooltip = 2,
		baselineLevel = 0,
	}
}

local function GetOptions()
	local options = {
		name = "Unhittable",
		type = "group",
		childGroups = "tab",
		get = function(info)
			return db[info[#info]]
		end,
		set = function(info, value)
			db[info[#info]] = value
			Unhittable:UpdateDisplay()
		end,
		args = {
			desc = {
				type = "description",
				name = L["Displays a variety of useful stats for tanking."].."\n",
				order = 1,
			},
			baselineLevel = {
				type = "range",
				name = L["Target Level (0 = Player Level)"],
				desc = L["Used for level difference in calculations."],
				min = 0, max = MAX_PLAYER_LEVEL + 3, step = 1,
				width = "full",
				order = 5,
			},
			display = {
				type = "group",
				name = L["Display"],
				desc = L["Control how values are to displayed."],
				order = 1,
				args = {
					header1 = { name = L["Data Feed"], type = "header", order = 10 },
					displayPrecisionFuBar = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed on the data feed."],
						min = 0, max = 4, step = 1,
						order = 11,
					},
					brokerLabel = {
						type = "select",
						name = L["Label Style"],
						desc = L["Style of labels to display on the data feed."],
						values = { [0] = L["None"], [1] = L["Short"], [2] = L["Long"] },
						order = 12,
					},
					showFuBarAV = {
						type = "toggle",
						name = L["Show Avoidance"],
						desc = L["Display the avoidance percentage on the data feed."],
						width = "full",
						order = 13,
					},
					showFuBarAVBlock = {
						type = "toggle",
						name = L["Include Block with Avoidance"],
						desc = L["Include block chance in the avoidance percentage on the data feed."],
						width = "full",
						hidden = function() return (PlayerClass ~= "WARRIOR" and PlayerClass ~= "PALADIN") end,
						disabled = function() return not db.showFuBarAV end,
						order = 14,
					},
					showFuBarDR = {
						type = "toggle",
						name = L["Show Damage Reduction"],
						desc = L["Display the damage reduction percentage on the data feed."],
						width = "full",
						order = 15,
					},
					showFuBarArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor value on the data feed."],
						width = "full",
						order = 16,
					},
					
					header2 = { name = L["Tooltip"], type = "header", order = 20 },
					displayPrecisionTooltip = {
						type = "range",
						name = L["Display Precision"],
						desc = L["Number of decimal places displayed in the tooltip."],
						min = 0, max = 4, step = 1,
						order = 21,
					},
					showTooltipCombat = {
						type = "toggle",
						name = L["Show Combat Table"],
						desc = L["Display the combat table values."],
						width = "full",
						order = 23,
					},
					showTooltipArmor = {
						type = "toggle",
						name = L["Show Armor"],
						desc = L["Display the armor values."],
						width = "full",
						order = 24,
					},
				},
			},
			color = {
				type = "group",
				name = L["Colors"],
				desc = L["Set the text colors."],
				order = 2,
				get = function(info)
					return unpack(db[info[#info]])
				end,
				set = function(info, r, g, b)
					db[info[#info]] = {r, g, b, 1}
					Unhittable:UpdateDisplay()
				end,
				args = {
					brokerHeader = {
						type = "header",
						name = L["Data Feed"],
						order = 10,
					},
					avoidColor = {
						type = "color",
						name = L["Avoidance"],
						desc = L["Set the color of the avoidance percentage on the data feed."],
						order = 11,
					},
					drColor = {
						type = "color",
						name = L["Damage Reduction"],
						desc = L["Set the color of the damage reduction percentage on the data feed."],
						order = 12,
					},
					armorColor = {
						type = "color",
						name = L["Armor"],
						desc = L["Set the color of the armor value on the data feed."],
						order = 13,
					},
					tooltipHeader = {
						type = "header",
						name = L["Tooltip"],
						order = 20,
					},
					headerColor = {
						type = "color",
						name = L["Header"],
						desc = L["Set the color of header text."],
						order = 21,
					},
					labelColor = {
						type = "color",
						name = L["Label"],
						desc = L["Set the color of label text."],
						order = 22,
					},
					valueColor = {
						type = "color",
						name = L["Value"],
						desc = L["Set the color of value text."],
						order = 23,
					},
				},
			},
		}
	}
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Unhittable.db)
	options.args.profile.order = 100
	return options
end


local GetMissChance
do
	--With Defense Rating removed, there is no longer a way to increase chance to be missed.
	local missChance = select(2, UnitRace("player")) == "NightElf" and 7 or 5
	function GetMissChance()
		return missChance
	end
end

local function GetArmorReduction(level)
	local armor = select(2, UnitArmor("player"))
	return PaperDollFrame_GetArmorReduction(armor, level or PlayerLevel)
end

local function GetCritReduction()
	local currentRank, maxRank = 0, 0
	if PlayerClass == "DEATHKNIGHT" then
		currentRank, maxRank = select(5, GetTalentInfo(1, 14))
	elseif PlayerClass == "DRUID" then
		currentRank, maxRank = select(5, GetTalentInfo(2, 18))
	elseif PlayerClass == "PALADIN" then
		currentRank, maxRank = select(5, GetTalentInfo(2, 8))
	elseif PlayerClass == "WARRIOR" then
		currentRank, maxRank = select(5, GetTalentInfo(3, 10))
	end

	return maxRank > 0 and (currentRank / maxRank * 6) or 0
end

local function GetCombatTable(level)
	if not level then level = PlayerLevel end

	local defenseDiff = (level - PlayerLevel) * 0.20
	local attackerDiff = level * 5 - PlayerLevel * 5

	local missChance = max(0, GetMissChance() - defenseDiff)
	local dodgeChance = max(0, GetDodgeChance() - defenseDiff)
	local parryChance = max(0, GetParryChance() - defenseDiff)
	local blockChance = max(0, GetBlockChance() - defenseDiff)
	local critChance = max(0, (5 + attackerDiff * 0.04) - GetCritReduction())
	local crushChance = 0
	if attackerDiff >= 20 then -- four level difference
		crushChance = max(0, attackerDiff * 0.02 - 0.15) * 100
	end
	local hitChance = 100 - min(100, missChance + dodgeChance + parryChance + blockChance + critChance + crushChance)

	return missChance, dodgeChance, parryChance, blockChance, critChance, crushChance, hitChance
end

local function GetHexColor(color)
	return format("%02x%02x%02x", color[1] * 255, color[2] * 255, color[3] * 255)
end

--StatGroup definition
local AddStatFrames
do
	local function CombatTableCoverage_OnEnter(statFrame)
		if MOVING_STAT_CATEGORY then return end

		GameTooltip:SetOwner(statFrame, "ANCHOR_RIGHT")
		GameTooltip:SetText(HIGHLIGHT_FONT_COLOR_CODE..format(PAPERDOLLFRAME_TOOLTIP_FORMAT, L["Combat Table Coverage"])..FONT_COLOR_CODE_CLOSE)
		GameTooltip:AddLine(L["Chance of dodging, parrying, blocking, or being missed by an attack."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, true)
		GameTooltip:AddLine(" ")

		GameTooltip:AddDoubleLine(STAT_TARGET_LEVEL, L["Avoidance"], HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		for i = 0, 3 do
			local level = PlayerLevel + i
			local missChance, dodgeChance, parryChance, blockChance = GetCombatTable(level)
			local total = missChance + dodgeChance + parryChance + blockChance
			local ctc
			if total <= 100 then
				ctc = format("%.2f%%", total)
			else
				ctc = format("100%% (+%.2f%%)", (total - 100))
			end

			if i == 3 then
				level = level.." / |TInterface\\TargetingFrame\\UI-TargetingFrame-Skull:0|t"
			end
			GameTooltip:AddDoubleLine("      "..level, ctc.."    ")
		end

		GameTooltip:Show()
	end

	local function SetCombatTableCoverage(statFrame, unit)
		if unit ~= "player" then
			statFrame:Hide()
			return
		end

		local missChance, dodgeChance, parryChance, blockChance = GetCombatTable()
		local ctc = missChance + dodgeChance + parryChance + blockChance
		PaperDollFrame_SetLabelAndText(statFrame, L["CTC"], ctc, true)
		statFrame:SetScript("OnEnter", CombatTableCoverage_OnEnter)
		statFrame:Show()
	end

	local function HitChance_OnEnter(statFrame)
		if MOVING_STAT_CATEGORY then return end

		GameTooltip:SetOwner(statFrame, "ANCHOR_RIGHT")
		GameTooltip:SetText(HIGHLIGHT_FONT_COLOR_CODE..format(PAPERDOLLFRAME_TOOLTIP_FORMAT, STAT_HIT_CHANCE)..FONT_COLOR_CODE_CLOSE)
		GameTooltip:AddLine(L["Chance of an attack hitting you."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, true)
		GameTooltip:AddLine(" ")

		GameTooltip:AddDoubleLine(STAT_TARGET_LEVEL, STAT_HIT_CHANCE, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		for i = 0, 3 do
			local level = PlayerLevel + i
			local hitChance = format("%.2f%%", select(7, GetCombatTable(level)))

			if i == 3 then
				level = level.." / |TInterface\\TargetingFrame\\UI-TargetingFrame-Skull:0|t"
			end
			GameTooltip:AddDoubleLine("      "..level, hitChance.."    ")
		end

		GameTooltip:Show()
	end

	local function SetHitChance(statFrame, unit)
		if unit ~= "player" then
			statFrame:Hide()
			return
		end

		local hitChance = select(7, GetCombatTable())
		PaperDollFrame_SetLabelAndText(statFrame, STAT_HIT_CHANCE, hitChance, true)
		statFrame:SetScript("OnEnter", HitChance_OnEnter)
		--statFrame.tooltip = format("|cffffffff%s %.2f%%|r", STAT_HIT_CHANCE, hitChance)
		--statFrame.tooltip2 = L["Chance of an attack hitting you."]
		statFrame:Show()
	end

	function AddStatFrames()
		table.insert(PAPERDOLL_STATCATEGORIES["DEFENSE"].stats, 5, "CTC")
		table.insert(PAPERDOLL_STATCATEGORIES["DEFENSE"].stats, 5, "CTC_HITCHANCE")
		PAPERDOLL_STATINFO["CTC_HITCHANCE"] = { updateFunc = SetHitChance }
		PAPERDOLL_STATINFO["CTC"] = { updateFunc = SetCombatTableCoverage }
	end
end


function Unhittable:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("UnhittableDB", defaults, true)
	db = self.db.profile

	local function OnProfileChanged()
		db = self.db.profile
		self:UpdateDisplay()
	end
	self.db.RegisterCallback(self, "OnProfileChanged", OnProfileChanged)
	self.db.RegisterCallback(self, "OnProfileCopied", OnProfileChanged)
	self.db.RegisterCallback(self, "OnProfileReset", OnProfileChanged)

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker Unhittable", GetOptions)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker Unhittable", "Broker Unhittable")

	ldb = LibStub("LibDataBroker-1.1"):NewDataObject("Broker_Unhittable", {
		type = "data source",
		icon = "Interface\\Icons\\Spell_Holy_DevotionAura",
		label = "Unhittable",
		text = "",
		OnClick = function(self, button)
			if button == "RightButton" then
				InterfaceOptionsFrame_OpenToCategory("Broker Unhittable")
			elseif button == "LeftButton" and (PlayerClass == "WARRIOR" or PlayerClass == "PALADIN") and db.showFuBarAV then
				db.showFuBarAVBlock = not db.showFuBarAVBlock
				Unhittable:UpdateDisplay()
			end
		end,
		OnTooltipShow = Unhittable.OnTooltipShow,
	})
end

function Unhittable:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateDisplay")
	self:RegisterEvent("COMBAT_RATING_UPDATE", "UpdateDisplay")
	self:RegisterEvent("MASTERY_UPDATE", "UpdateDisplay")
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "UpdateDisplay")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateDisplay")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED", "UpdateDisplay")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterBucketEvent("UNIT_AURA", 1, "BucketUpdate")

	AddStatFrames()
end

function Unhittable:PLAYER_LEVEL_UP(_, level)
	PlayerLevel = level
	self:UpdateDisplay()
end

function Unhittable:BucketUpdate(units)
	if units.player then
		self:UpdateDisplay()
	end
end

function Unhittable:UpdateDisplay()
	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or PlayerLevel

	local displayString = "|r"
	local formatString = format("%%s|cff%%s%%.%df%%%%|r ", db.displayPrecisionFuBar)
	local label = db.brokerLabel

	if db.showFuBarAV then
		local missChance, dodgeChance, parryChance, blockChance = GetCombatTable(calcLevel)
		local av = missChance + dodgeChance + parryChance

		if label == 1 then
			displayString = format("%s%s", displayString, L["Av:"])
		elseif label == 2 then
			displayString = format("%s%s", displayString, L["Avoidance:"])
		end
		displayString = formatString:format(displayString, GetHexColor(db.avoidColor), db.showFuBarAVBlock and av + blockChance or av)
	end

	if db.showFuBarDR then
		if label == 1 then
			displayString = format("%s%s", displayString, L["DR:"])
		elseif label == 2 then
			displayString = format("%s%s", displayString, L["Damage Reduction:"])
		end
		displayString = formatString:format(displayString, GetHexColor(db.drColor), GetArmorReduction(calcLevel))
	end

	if db.showFuBarArmor then
		local armor = select(2, UnitArmor("player"))

		if label == 1 then
			displayString = format("%s%s", displayString, L["Ar:"])
		elseif label == 2 then
			displayString = format("%s%s", displayString, L["Armor:"])
		end
		displayString = format("%s|cff%s%d|r ", displayString, GetHexColor(db.armorColor), armor)
	end  

	ldb.text = displayString
end

function Unhittable.OnTooltipShow(tooltip)
	if not tooltip or not tooltip.AddLine then return end

	local calcLevel = db.baselineLevel > 0 and db.baselineLevel or PlayerLevel
	local missChance, dodgeChance, parryChance, blockChance, critChance, crushChance, hitChance = GetCombatTable(calcLevel)

	local formatString = format("%%.%df%%%%", db.displayPrecisionTooltip)
	local hr, hg, hb = unpack(db.headerColor)
	local lr, lg, lb = unpack(db.labelColor)
	local vr, vg, vb = unpack(db.valueColor)

	tooltip:AddLine(HIGHLIGHT_FONT_COLOR_CODE .. "Unhittable")
	tooltip:AddDoubleLine(L["Total (vs lv%d)"]:format(calcLevel), formatString:format(missChance + dodgeChance + parryChance + blockChance), lr, lg, lb, vr, vg, vb)

	if db.showTooltipCombat then
		tooltip:AddLine(" ")
		tooltip:AddLine(L["Combat Table"], hr, hg, hb)
		tooltip:AddDoubleLine(L["Miss Chance"], formatString:format(missChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Dodge Chance"], formatString:format(dodgeChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Parry Chance"], formatString:format(parryChance), lr, lg, lb, vr, vg, vb)
		tooltip:AddDoubleLine(L["Block Chance"], formatString:format(blockChance), lr, lg, lb, vr, vg, vb)
		if critChance > 0 then
			tooltip:AddDoubleLine(L["Crit Chance"], formatString:format(critChance), lr, lg, lb, vr, vg, vb)
		end
		if crushChance > 0 then
			tooltip:AddDoubleLine(L["Crush Chance"], formatString:format(crushChance), lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Hit Chance"], formatString:format(hitChance), lr, lg, lb, vr, vg, vb)
	end

	if db.showTooltipArmor then
		local baseArmor, effectiveArmor, _, armorBuffs, armorDebuffs = UnitArmor("player")

		tooltip:AddLine(" ")
		tooltip:AddLine(L["Armor"], hr, hg, hb)
		if baseArmor == effectiveArmor then
			tooltip:AddDoubleLine(L["Armor"], effectiveArmor, lr, lg, lb, vr, vg, vb)
		else
			tooltip:AddDoubleLine(L["Base Armor"], baseArmor, lr, lg, lb, vr, vg, vb)
			if armorBuffs > 0 then
				tooltip:AddDoubleLine(L["Armor Buffs"], "+"..armorBuffs, lr, lg, lb, vr, vg, vb)
			end
			if armorDebuffs < 0 then
				tooltip:AddDoubleLine(L["Armor Debuffs"], "-"..armorDebuffs, lr, lg, lb, vr, vg, vb)
			end
			tooltip:AddDoubleLine(L["Total Armor"], effectiveArmor, lr, lg, lb, vr, vg, vb)
		end
		tooltip:AddDoubleLine(L["Damage Reduction"], formatString:format(GetArmorReduction(calcLevel)), lr, lg, lb, vr, vg, vb)
	end
end